.. Global includes, substitutions, and common links.

.. |author_name| replace:: Yesudeep Mangalapilly
.. |author_email| replace:: yesudeep@gmail.com
.. |copyright| replace:: Copyright 2012 Google, Inc & contributors.
.. |project_name| replace:: ``watchdog``
.. |project_version| replace:: 2.1.4

.. _issue tracker: https://github.com/gorakhargosh/watchdog/issues
.. _code repository: https://github.com/gorakhargosh/watchdog
.. _mailing list: https://groups.google.com/group/watchdog-python

.. _kqueue: https://www.freebsd.org/cgi/man.cgi?query=kqueue&sektion=2
.. _FSEvents: https://developer.apple.com/library/mac/#documentation/Darwin/Conceptual/FSEvents_ProgGuide/Introduction/Introduction.html
.. _inotify: https://linux.die.net/man/7/inotify
.. _Mac OS X File System Monitoring Performance Guidelines: https://developer.apple.com/library/ios/#documentation/Performance/Conceptual/FileSystem/Articles/TrackingChanges.html
.. _ReadDirectoryChangesW: https://docs.microsoft.com/windows/win32/api/winbase/nf-winbase-readdirectorychangesw
.. _GetQueuedCompletionStatus: https://docs.microsoft.com/windows/win32/api/ioapiset/nf-ioapiset-getqueuedcompletionstatus
.. _CreateIoCompletionPort: https://docs.microsoft.com/windows/win32/api/ioapiset/nf-ioapiset-createiocompletionport

.. _argh: https://pypi.python.org/pypi/argh
.. _coverage: https://coverage.readthedocs.io/en/latest/
.. _file.monitor: https://github.com/pke/file.monitor
.. _fsmonitor: https://github.com/shaurz/fsmonitor
.. _git: https://git-scm.org/
.. _github: https://github.com/
.. _guard: https://github.com/guard/guard
.. _homebrew: https://brew.sh/
.. _inotify-tools: https://github.com/rvoicilas/inotify-tools
.. _jnotify: http://jnotify.sourceforge.net/
.. _LibYAML: https://pyyaml.org/wiki/LibYAML
.. _nose: https://nose.readthedocs.org/en/latest/
.. _pip: https://pypi.python.org/pypi/pip
.. _pnotify: http://mark.heily.com/pnotify
.. _pyfilesystem: https://github.com/PyFilesystem/pyfilesystem
.. _pyinotify: https://github.com/seb-m/pyinotify
.. _Python: https://python.org
.. _pywin32: https://github.com/mhammond/pywin32
.. _PyYAML: https://www.pyyaml.org/
.. _sphinx: https://www.sphinx-doc.org/en/master/
.. _treewalker: https://github.com/jbd/treewatcher
.. _unison fsmonitor: https://webdav.seas.upenn.edu/viewvc/unison/trunk/src/fsmonitor.py?view=markup&pathrev=471
.. _XCode: https://developer.apple.com/technologies/tools/xcode.html
.. _zc.buildout: https://www.buildout.org/
